#version 460

layout(r11f_g11f_b10f) uniform image3D lighting_map;
layout(rgba8) uniform image2DArray voxelModels;

uniform usamplerBuffer voxelMasks;
uniform mat4 perspective;
uniform mat4 model;
uniform mat4 view;

in vec3 uv;
in vec2 viewport;
in flat uint voxelTexture;
in flat ivec3 voxelPosition;
in flat ivec3 faceMask;

out vec4 color;

void main() {
	vec3 direction = (inverse(perspective) * vec4(viewport, 0.0f, 1.0f)).xyz;

	direction = normalize((inverse(view * model) * vec4(direction, 0.0f)).xyz);

	ivec3 rayStep = ivec3(sign(direction));
	vec3 ray_pos = vec3(uv * 8.0 + direction * 0.01);
	ivec3 mapPos = ivec3(floor(ray_pos));

	float dirLength = length(direction);
	vec3 deltaDist = abs(dirLength.xxx / direction);
	vec3 sideDist = (sign(direction) * (vec3(mapPos) - ray_pos) + (sign(direction) * 0.5) + 0.5) * deltaDist;

	float hitDistance = 0;
	ivec3 mask = faceMask;

	while (all(greaterThanEqual(mapPos, ivec3(0))) && all(lessThan(mapPos, ivec3(8))))
	{
		uint index = mapPos.x + mapPos.y * 8 + (7 - mapPos.z) * 64;
		if (((texelFetch(voxelMasks, int(voxelTexture * 16 + index / 32)).r >> (index % 32)) & 1) == 1) {
			color = imageLoad(voxelModels, ivec3((mapPos.x + mapPos.y * 8), mapPos.z, voxelTexture));

			ivec3 size = imageSize(lighting_map);
			if (voxelPosition.x < 0 || voxelPosition.y < 0 || voxelPosition.z < 0 || voxelPosition.x >= size.x || voxelPosition.y >= size.y || voxelPosition.z >= size.z) {
				color.rgb *= 0.1;
			}
 			else {
				ivec3 normal = rayStep * mask;
				
				if (normal.x == 1) {
					color.rgb *= imageLoad(lighting_map, voxelPosition - ivec3(1, 0, 0)).rgb * 0.8;
				} else if (normal.x == -1) {
					color.rgb *= imageLoad(lighting_map, voxelPosition + ivec3(1, 0, 0)).rgb * 0.8;
				} else if (normal.y == 1) {
					color.rgb *= imageLoad(lighting_map, voxelPosition - ivec3(0, 1, 0)).rgb * 1.0;
				} else if (normal.y == -1) {
					color.rgb *= imageLoad(lighting_map, voxelPosition + ivec3(0, 1, 0)).rgb * 0.4;
				} else if (normal.z == 1) {
					color.rgb *= imageLoad(lighting_map, voxelPosition - ivec3(0, 0, 1)).rgb * 0.6;
				} else if (normal.z == -1) {
					color.rgb *= imageLoad(lighting_map, voxelPosition + ivec3(0, 0, 1)).rgb * 0.6;
				}
			}
			return;
		}

		hitDistance = min(min(sideDist.x, sideDist.y), sideDist.z);
		mask = ivec3(sideDist.x == hitDistance, sideDist.y == hitDistance, sideDist.z == hitDistance);
		sideDist += mask * deltaDist;
		mapPos += mask * rayStep;
	}

	discard;
}