#version 450
	
uniform mat4 perspective;
uniform mat4 view;
uniform mat4 model;
uniform ivec3 light_offset;

in uint data;

out vec3 uv;
out uint voxelTexture;
out vec2 viewport;
out ivec3 voxelPosition;
out ivec3 faceMask;

// Offsets for our vertices drawing this face
const vec3 facePositions[24] = vec3[24]
(
	// East
	vec3(0.5f, -0.5f, -0.5f),
	vec3(0.5f, 0.5f, -0.5f),
	vec3(0.5f, 0.5f, 0.5f),
	vec3(0.5f, -0.5f, 0.5f),

	// Up
	vec3(-0.5f, 0.5f, -0.5f),
	vec3(-0.5f, 0.5f, 0.5f),
	vec3(0.5f, 0.5f, 0.5f),
	vec3(0.5f, 0.5f, -0.5f),

	// North
	vec3(0.5f, -0.5f, 0.5f),
	vec3(0.5f, 0.5f, 0.5f),
	vec3(-0.5f, 0.5f, 0.5f),
	vec3(-0.5f, -0.5f, 0.5f),

	// West
	vec3(-0.5f, -0.5f, 0.5f),
	vec3(-0.5f, 0.5f, 0.5f),
	vec3(-0.5f, 0.5f, -0.5f),
	vec3(-0.5f, -0.5f, -0.5f),

	// Down
	vec3(-0.5f, -0.5f, 0.5f),
	vec3(-0.5f, -0.5f, -0.5f),
	vec3(0.5f, -0.5f, -0.5f),
	vec3(0.5f, -0.5f, 0.5f),

	// South
	vec3(-0.5f, -0.5f, -0.5f),
	vec3(-0.5f, 0.5f, -0.5f),
	vec3(0.5f, 0.5f, -0.5f),
	vec3(0.5f, -0.5f, -0.5f)
);

const vec3 faceUvs[24] = vec3[24](
	// East
	vec3(1, 0, 0),
	vec3(1, 1, 0),
	vec3(1, 1, 1),
	vec3(1, 0, 1),

	// Up
	vec3(0, 1, 0),
	vec3(0, 1, 1),
	vec3(1, 1, 1),
	vec3(1, 1, 0),
	
	// North
	vec3(1, 0, 1),
	vec3(1, 1, 1),
	vec3(0, 1, 1),
	vec3(0, 0, 1),

	// West
	vec3(0, 0, 1),
	vec3(0, 1, 1),
	vec3(0, 1, 0),
	vec3(0, 0, 0),

	// Down
	vec3(0, 0, 1),
	vec3(0, 0, 0),
	vec3(1, 0, 0),
	vec3(1, 0, 1),

	// South
	vec3(0, 0, 0),
	vec3(0, 1, 0),
	vec3(1, 1, 0),
	vec3(1, 0, 0)
);

const vec3 faceNormals[6] = vec3[6](
	// East
	vec3(1, 0, 0),

	// Up
	vec3(0, 1, 0),
	
	// North
	vec3(0, 0, 1),

	// West
	vec3(-1, 0, 0),

	// Down
	vec3(0, -1, 0),

	// South
	vec3(0, 0, -1)
);

const vec3 colorMap[2] = vec3[2]
(
	vec3(1.0f, 1.0f, 1.0f),
	vec3(0.3f, 0.3f, 0.3f)
);

void main()
{
	// Unpack the data we retrieved
	const uint x = (data) & 31; // 0-31 x position
	const uint y = (data >> 5) & 31; // 0-31 y position
	const uint z = (data >> 10) & 31; // 0-31 z position

	voxelTexture = (data >> 15); // 17 bits, texture index

	// Calculate real positions
	voxelPosition = light_offset + ivec3(x, y, z);
	vec3 position = (vec3(x,y,z) + facePositions[gl_VertexID]);

	faceMask = ivec3(0);
	faceMask[(gl_VertexID / 4) % 3] = 1;

	// Output our uv
	uv = faceUvs[gl_VertexID];

	// Output our position
	gl_Position = perspective * view * model * vec4(position, 1.0);
	viewport = gl_Position.xy / gl_Position.w;
}